/*\
|*| ---------------------------------------------------------------------------
|*|	Cirrus Logic programable clocks
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Martin Erhardsen, [u930534@daimi.aau.dk].
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*| ---------------------------------------------------------------------------
|*|	This code was derived from vgadoc4b.zip and sdac.c.
|*|	
\*/

#define	MAINTAINER	"Martin Erhardsen, u930534@daimi.aau.dk"

#include <linux/kernel.h>
#define DEBUG
#include <ggi/graphdev.h>

struct kgi_clock clock =
{
	"Cirrus Logic",	        /* manufacturer		*/ 
	"542x",	                /* model		*/
	GM_ALL,			/* supported modes 	*/
	0,			/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/

	CLK_TYPE_PROG,		/* clock type		*/
	{{   400000,  86000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
  };


#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

/*\
|*|	Clock driver dependend #define's
\*/

#define CLK_VERSION "0.00 ALPHA"

#define	fref	14318180   /* Hz, must be within 14..15 MHz */ 

#define F(M,N,FVCO)  (((M)*(FVCO))/(N))

#define	INVALIDCLK 0xFFFF

static unsigned short clk_last_cmd = 0xFFFF;
static int 	      clk_last_freq = -1;


unsigned short
calcfreq(int *freq, enum kgi_tm_cmd cmd)
{
  int fvco = *freq;
  int hf   = *freq;
  int div  = 0;
  int n;
  struct { unsigned char m; unsigned char n; int delta; }
         best = { 0xFF, 0xFF, 0x7FFFFFFF };    
 
  if(*freq == clk_last_freq)	/* we never get called with freq == -1 */
    return clk_last_cmd;
  
  if (hf < 25000000)
    { div = 1; hf += hf; };
 
  fvco  = fref >> div;

  for(n = 1; n < 32; n++)
    { 
      register int m = (hf*n)/fref;
      register int delta = *freq - F(m,n,fvco);

      delta = (cmd == CMD_LOWER) ? delta : -delta;

      if(m > 127) break;
      if((delta < 0) || (delta > best.delta)) continue;

      best.n = n; best.m = m; best.delta = delta;
    }

  if((best.n > 31) || (best.m > 127) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz), n=%i, m=%i, d=%i\n",
                   __FILE__,__LINE__, *freq, best.n, best.m, div)); 
      return INVALIDCLK; 
    }
   else
    {
      TRACE(printk("%s: request for DCLK = %i Hz,", __FILE__,*freq ));
      *freq = clk_last_freq = F(best.m,best.n,fref >> div);
      TRACE(printk(" did %i Hz.\n", *freq));
      return clk_last_cmd = (best.m<<8) | (best.n << 1) | div;
    }
}

/*\
|*|     GGI clock driver interface
\*/

int
kgi_clock_init(void)
{
  printk("%s %s clock driver v"CLK_VERSION".\n", clock.manufact, clock.model);
  return EOK;
}

void
kgi_clock_done(void)
{
  TRACE(printk("%s %s clock driver removed.\n", clock.manufact, clock.model));
}

struct kgi_clock
*kgi_clock_get_info (void)
{ 
  return &clock;
}

int
kgi_clock_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newfreq = (cmd == CMD_PROPOSE) 
                    ? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->clock = INVALIDCLK;
	return EOK;

     case CMD_LOWER:
	if (newfreq < MIN_FREQ)
	  {
	    TRACE(printk("%s: requested DCLK too low (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -ECLOCK;
	  }

   	if (newfreq > MAX_FREQ)	/* meet limits */
	  newfreq = MAX_FREQ;

	mode->clock = calcfreq(&newfreq, cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;
			
     case CMD_RAISE:
	if (MAX_FREQ < newfreq)
	  {
	    TRACE(printk("%s: requested DCLK too high (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -ECLOCK;
	  }

	if (newfreq < MIN_FREQ)
	  newfreq = MIN_FREQ;

	mode->clock = calcfreq(&newfreq, cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;

     case CMD_CHECK:
	if ((newfreq < MIN_FREQ) || (newfreq > MAX_FREQ))
	  return -ECLOCK;

	mode->clock = calcfreq(&newfreq, CMD_LOWER);

	if (   (mode->clock != INVALIDCLK)
	    && (   ((mode->dclk*mode->clk.mul/mode->clk.div) - newfreq) 
		 < ((mode->dclk*mode->clk.mul/mode->clk.div)*1/100)))
	  return EOK;

	return -ECLOCK;

     default:
	INTERNALERROR;
	return -ECLOCK;
    }
}

int
kgi_clock_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg)
{
  return -ECLOCK;
}
